@@ -2,9 +2,10 @@
 import abc
 import collections
 import fnmatch
+import itertools
 import re
-import urllib.parse
 import sys
+import urllib.parse
 
 
 URLInfoType = collections.namedtuple(
@@ -58,12 +59,12 @@ class URLInfo(URLInfoType):
         return URLInfo(
             url_split_result.scheme,
             url_split_result.netloc,
-            url_split_result.path or '/',
-            url_split_result.query,
+            cls.normalize_path(url_split_result.path),
+            cls.normalize_query(url_split_result.query),
             url_split_result.fragment,
             url_split_result.username,
             url_split_result.password,
-            url_split_result.hostname,
+            cls.normalize_hostname(url_split_result.hostname),
             port,
             string,
             cls.normalize(url_split_result),
@@ -77,18 +78,76 @@ class URLInfo(URLInfoType):
         default_port = cls.DEFAULT_PORTS.get(url_split_result.scheme)
 
         if default_port == url_split_result.port:
-            host_with_port = url_split_result.hostname
+            host_with_port = cls.normalize_hostname(url_split_result.hostname)
         else:
             host_with_port = url_split_result.netloc.split('@', 1)[-1]
+            if ':' in host_with_port:
+                host, port = host_with_port.rsplit(':', 1)
+                host_with_port = '{0}:{1}'.format(
+                    cls.normalize_hostname(host), port)
+            else:
+                host_with_port = cls.normalize_hostname(host_with_port)
 
         return urllib.parse.urlunsplit([
             url_split_result.scheme,
             host_with_port,
-            url_split_result.path or '/',
-            url_split_result.query,
+            cls.normalize_path(url_split_result.path),
+            cls.normalize_query(url_split_result.query),
             b''
         ])
 
+    @classmethod
+    def normalize_hostname(cls, hostname):
+        if hostname:
+            return hostname.encode('idna').decode('ascii')
+        else:
+            return hostname
+
+    @classmethod
+    def normalize_path(cls, path):
+        if path is None:
+            return
+
+        # First assume the path is quoted
+        unquoted_path = unquote(path)
+
+        # Then quote it back
+        requoted_path = quote(unquoted_path)
+
+        # If they match, they are indeed quoted correctly:
+        if uppercase_percent_encoding(path) == requoted_path:
+            return path or '/'
+
+        # Otherwise, we'll assume that it's unquoted
+        return quote(path) or '/'
+
+    @classmethod
+    def normalize_query(cls, query):
+        if query is None:
+            return
+
+        query_list = split_query(query, keep_blank_values=True)
+        query_test_str = ''.join(itertools.chain(*query_list))
+
+        # First assume the query is quoted
+        unquoted_query_test_str = unquote_plus(query_test_str)
+
+        # Then quote it back
+        requoted_query_test_str = quote_plus(unquoted_query_test_str)
+
+        # If they match, they are indeed quoted correctly:
+        if uppercase_percent_encoding(query_test_str) \
+        == requoted_query_test_str:
+            return uppercase_percent_encoding(query)
+
+        # Otherwise, we'll assume that it's unquoted
+        return '&'.join([
+            '='.join((
+                quote_plus(name),
+                quote_plus(value)
+            ))
+            for name, value in query_list])
+
     def to_dict(self):
         return {
             'scheme': self.scheme,
@@ -290,3 +349,71 @@ def is_subdir(base_path, test_path, trailing_slash=False, wildcards=False):
         return fnmatch.fnmatchcase(test_path, base_path)
     else:
         return test_path.startswith(base_path)
+
+
+def quote(string, safe='/', encoding='utf-8', errors='strict'):
+    if sys.version_info[0] == 2:
+        # Backported behavior
+        return urllib.parse.quote(
+            string.encode(encoding, errors),
+            safe.encode(encoding, errors)
+        ).decode(encoding, errors)
+    else:
+        return urllib.parse.quote(string, safe, encoding, errors)
+
+
+def quote_plus(string, safe='', encoding='utf-8', errors='strict'):
+    if sys.version_info[0] == 2:
+        # Backported behavior
+        return urllib.parse.quote_plus(
+            string.encode(encoding, errors),
+            safe.encode(encoding, errors)
+        ).decode(encoding, errors)
+    else:
+        return urllib.parse.quote_plus(string, safe, encoding, errors)
+
+
+def unquote(string, encoding='utf-8', errors='strict'):
+    if sys.version_info[0] == 2:
+        return urllib.parse.unquote(
+            string.encode(encoding, errors)
+        ).decode(encoding, errors)
+    else:
+        return urllib.parse.unquote(string, encoding, errors)
+
+
+def unquote_plus(string, encoding='utf-8', errors='strict'):
+    if sys.version_info[0] == 2:
+        return urllib.parse.unquote_plus(
+            string.encode(encoding, errors)
+        ).decode(encoding, errors)
+    else:
+        return urllib.parse.unquote_plus(string, encoding, errors)
+
+
+def split_query(qs, keep_blank_values=False):
+    new_list = []
+
+    for pair in qs.split('&'):
+        items = pair.split('=', 1)
+
+        if len(items) == 0:
+            continue
+
+        if len(items) == 1:
+            name = items[0]
+            value = ''
+        else:
+            name, value = items
+
+        if keep_blank_values or value:
+            new_list.append((name, value))
+
+    return new_list
+
+
+def uppercase_percent_encoding(string):
+    return re.sub(
+        r'%[a-f0-9][a-f0-9]',
+        lambda match: match.group(0).upper(),
+        string)
